#include <boost/shared_ptr.hpp>
#include <fstream>
#include <iomanip>
#include <iostream>
#include <mimas/mimasexception.h>
#include <mimas/gauss.h>
#include <mimas/image.h>
#include <mimas/image_fileinput.h>
#include <mimas/image_fileoutput.h>
#include <mimas/image_mesaoutput.h>
#include <mimas/multi_array_op.h>
#include <mimas/rgba.h>
#include <popt.h>
#include <string>
#include <vector>
#include "tools.h"

// #define X11OUTPUT

using namespace boost;
using namespace mimas;
using namespace std;

template< typename T >
struct _sqr: public unary_function< T, T > 
{
  T operator()( const T &x ) const { return x * x; }
};

template< typename T >
struct _log: public unary_function< T, T >
{
  T operator()( const T &x ) const { return log( x ); }
};

template< typename T >
struct _todouble: public unary_function< T, double > {
  double operator()( const T &x ) const
  { return static_cast< double >( x ); }
};

template< typename T >
struct _touchar: public unary_function< T, unsigned char > {
  unsigned char operator()( const T &x ) const
  { return static_cast< unsigned char >( x ); }
};

ARRAY_OP_EX( sqr , _sqr )
ARRAY_OP_EX( log , _log )
ARRAY_OP_EX( todouble, _todouble )
ARRAY_OP_EX( touchar, _touchar )

int main( int argc, char *argv[] )
{
  int retVal = 0;

  int alternativeDeepview = 0;

  double sigma = 5.0;
  const char
    *heightFieldFileName = NULL,
    *deepViewFileName = NULL;

  struct poptOption optionsTable[] = {
    { "sigma", 's', POPT_ARG_DOUBLE, &sigma, 0,
      "sigma for gaussian blur (default 5)", "sigma/pixelsize" },
    { "heightfield", 'h', POPT_ARG_STRING, &heightFieldFileName, 0,
      "output pgm-file name for height field", "ppm file-name" },
    { "deepview", 'd', POPT_ARG_STRING, &deepViewFileName, 0,
      "output ppm-file name for deep view", "ppm file-name" },
    { "alternative", 'a', POPT_ARG_VAL, &alternativeDeepview, 1,
      "threat later half of filenames as alternative focus stack to generate deep view.", NULL },
    POPT_AUTOHELP
    { NULL, 0, 0, NULL, 0 } };
  
  poptContext optCon( poptGetContext( NULL, argc, (const char **)argv,
                                      optionsTable, 0 ) );
  poptSetOtherOptionHelp( optCon, "images ..." );

  try {

    char c;
    while ( ( c = poptGetNextOpt( optCon ) ) >= 0 );

    MMERROR( c >= -1, mimasexception, ,
             poptBadOption( optCon, POPT_BADOPTION_NOALIAS ) << ": "
             << poptStrerror( c ) );

    MMERROR( heightFieldFileName != NULL, mimasexception, ,
             "No file name for height field was specified." );

    MMERROR( deepViewFileName != NULL, mimasexception, ,
             "No file name for deep view was specified." );

    vector< string > commandLineFileNames;
    const char *fileName;
    while ( ( fileName = poptGetArg( optCon ) ) != NULL )
      commandLineFileNames.push_back( fileName );

    vector< string > fileNames, alternativeFileNames;
    if ( alternativeDeepview != 0 ) {
      MMERROR( commandLineFileNames.size() % 2 == 0, mimasexception, ,
               "Number of filenames must be even, when specifying two focus stacks." );
      for ( int i=0; i<(signed)commandLineFileNames.size() / 2; i++ )
        fileNames.push_back( commandLineFileNames[i] );
      for ( int i=(signed)commandLineFileNames.size() / 2; i<(signed)commandLineFileNames.size();
            i++ )
        alternativeFileNames.push_back( commandLineFileNames[i] );
    } else {
      fileNames = commandLineFileNames;
      alternativeFileNames = commandLineFileNames;
    };

    MMERROR( fileNames.size() <= 255, mimasexception, ,
             "Cannot handle more than 255 files." );

    shared_ptr< multi_array< unsigned char, 2 > > height;
    shared_ptr< multi_array< double, 2 > > maxSharpness;

#ifdef X11OUTPUT
    x11_display display;
    image_mesaoutput< unsigned char > win( &display );
#endif

    for ( unsigned char i=0; i<fileNames.size(); i++ ) {
      ifstream file( fileNames[i].c_str(), ios::binary );
      image< double > image;
      MMERROR( file >> image, mimasexception, ,
               "Error loading file \"" << fileNames[i] << "\"." );
      int
        w = image.getWidth(),
        h = image.getHeight();

      if ( !height ) {
        height =
          shared_ptr< multi_array< unsigned char, 2 > >
          ( new multi_array< unsigned char, 2 >( extents[ h - 2 ][ w - 2 ] ) );
        maxSharpness =
          shared_ptr< multi_array< double, 2 > >
          ( new multi_array< double, 2 >( extents[ h - 2 ][ w - 2 ] ) );
      } else
        MMERROR( (signed)height->shape()[1] == w - 2 &&
                 (signed)height->shape()[0] == h - 2,
                 mimasexception, , "Images have different dimensions." );

      multi_array< double, 2 > data( image_to_multi_array( image ) );

      /* multi_array< double, 2 >
        sharpness( gauss( sqr( data ) ) - sqr( gauss( data ) ) ); */

      typedef multi_array< double, 2 >::index_range range;
      typedef multi_array< double, 2 >::const_array_view< 2 >::type array_view;
      multi_array< double, 2 >::index_gen indices;
      multi_array< double, 2 >
        sharpness( gaussian_blur
                   ( (const const_multi_array_ref< double, 2 > &)
                     ( sqr( data[ indices[ range( 2, h     ) ]
                                  [ range( 1, w - 1 ) ] ] -
                            data[ indices[ range( 0, h - 2 ) ]
                                  [ range( 1, w - 1 ) ] ] ) * 0.5 +
                       sqr( data[ indices[ range( 1, h - 1 ) ]
                                  [ range( 2, w     ) ] ] -
                            data[ indices[ range( 1, h - 1 ) ]
                                  [ range( 0, w - 2 ) ] ] ) * 0.5 ),
                     sigma, 0.1 ) );
      const double *p = sharpness.data();
      double *q = maxSharpness->data();
      unsigned char *r = height->data();
      for ( int j=0; j<(signed)sharpness.num_elements(); j++ ) {
        if ( *p > *q ) {
          *q = *p;
          *r = i;
        };
        p++;
        q++;
        r++;
      };
      /*
      sharpness = log( sharpness + 1.0 );
      window << multi_array_to_image
        ( sharpness *
          ( 255.0 /
            *max_element( sharpness.data(),
                          sharpness.data() + sharpness.num_elements() ) ) );
      */
      cerr << "*" << flush;
    };
    cerr << endl;
    
    *height = touchar( gaussian_blur( (const const_multi_array_ref< double, 2 > &)
                                      ( todouble( *height ) * *maxSharpness ),
                                      sigma, 0.1 ) /
                       gaussian_blur( (const const_multi_array_ref< double, 2 > &)
                                      ( *maxSharpness ), sigma, 0.1 ) );

#ifdef X11OUTPUT
    window << multi_array_to_image( *height *
				    (unsigned char)( 255.0 / fileNames.size() ) );
#endif
    multi_array< rgba< unsigned char >, 2 >
      deepView( extents[ height->shape()[0] ][ height->shape()[1] ] );

    for ( unsigned char i=0; i<alternativeFileNames.size(); i++ ) {
      ifstream file( alternativeFileNames[i].c_str(), ios::binary );
      image< rgba< unsigned char > > image;
      MMERROR( file >> image, mimasexception, ,
               "Error loading file \"" << fileNames[i] << "\"." );

      typedef multi_array< rgba< unsigned char >, 2 >::index_range range;
      typedef multi_array< rgba< unsigned char >, 2 >::
        const_array_view< 2 >::type array_view;
      multi_array< rgba< unsigned char >, 2 >::index_gen indices;
      
      multi_array< rgba< unsigned char >, 2 >
        array( image_to_multi_array( image ) );

      array_view
        view( array[ indices[ range( 1, height->shape()[0] + 1 ) ]
                            [ range( 1, height->shape()[1] + 1 ) ] ] );

      multi_array< rgba< unsigned char >, 2 >
        data( extents[ height->shape()[0] ][ height->shape()[1] ] );

      data = view;

      const rgba< unsigned char > *p = data.data();
      const unsigned char *q = height->data();
      rgba< unsigned char > *r = deepView.data();
      for ( int j=0; j<(signed)data.num_elements(); j++ ) {
        if ( *q == i )
          *r = *p;
        p++; q++; r++;
      };
    };

    image< rgba< unsigned char > > deepViewImage
      ( multi_array_to_image( deepView ) );

    ofstream deepViewFile( deepViewFileName, ios::binary );
    deepViewFile << deepViewImage;
    deepViewFile.close();

    ofstream heightFieldFile( heightFieldFileName, ios::binary );
    heightFieldFile << multi_array_to_image( *height );
    heightFieldFile.close();

#ifdef X11OUTPUT
    image_mesaoutput< unsigned char > win2( &display );

    *maxSharpness = log( *maxSharpness + 1.0 );
    win2 << multi_array_to_image
      ( *maxSharpness *
        ( 255.0 / *max_element( maxSharpness->data(),
                                maxSharpness->data() +
                                maxSharpness->num_elements() ) ) );

    image_qtoutput< rgba< unsigned char > > win3( &display );
    win3 << deepViewImage << mimas::pause();
#endif
    
  } catch ( std::exception &e ) {
    cerr << e.what() << endl;
    poptPrintUsage( optCon, stderr, 0 );
    retVal = 1;
  };
  poptFreeContext( optCon );
  return retVal;
}
